@@ -23,6 +23,7 @@ class Base(SourceNode, ABC):
     _type: Optional[Type.Base] = None
     _check_quant: bool = True
     _stdlib: "Optional[StdLib.Base]" = None
+    _struct_types: Optional[Env.Bindings[Dict[str, Type.Base]]] = None
 
     @property
     def type(self) -> Type.Base:
@@ -46,10 +47,11 @@ class Base(SourceNode, ABC):
     def infer_type(
         self,
         type_env: Env.Bindings[Type.Base],
-        stdlib: "Optional[StdLib.Base]" = None,
+        stdlib: StdLib.Base,
         check_quant: bool = True,
+        struct_types: Optional[Env.Bindings[Dict[str, Type.Base]]] = None,
     ) -> "Base":
-        """infer_type(self, type_env : Env.Bindings[Type.Base]) -> WDL.Expr.Base
+        """infer_type(self, type_env : Env.Bindings[Type.Base], stdlib : StdLib.Base) -> WDL.Expr.Base
 
         Infer the expression's type within the given type environment. Must be
         invoked exactly once prior to use of other methods.
@@ -69,13 +71,21 @@ class Base(SourceNode, ABC):
         with Error.multi_context() as errors:
             for child in self.children:
                 assert isinstance(child, Base)
-                errors.try1(lambda child=child: child.infer_type(type_env, stdlib, check_quant))
-        # invoke derived-class logic. we pass check_quant and stdlib hackily
-        # through instance variables since only some subclasses use them.
+                errors.try1(
+                    lambda child=child: child.infer_type(
+                        type_env, stdlib, check_quant, struct_types
+                    )
+                )
+        # invoke derived-class logic. we pass check_quant, stdlib, and struct_types hackily through
+        # instance variables since only some subclasses use them.
         self._check_quant = check_quant
         self._stdlib = stdlib
-        self._type = self._infer_type(type_env)
-        self._stdlib = None
+        self._struct_types = struct_types
+        try:
+            self._type = self._infer_type(type_env)
+        finally:
+            self._stdlib = None
+            self._struct_types = None
         assert self._type and isinstance(self.type, Type.Base)
         return self
 
@@ -93,16 +103,12 @@ class Base(SourceNode, ABC):
         return self
 
     @abstractmethod
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         # to be overridden by subclasses. eval() calls this and deals with any
         # exceptions raised
         pass
 
-    def eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         """
         Evaluate the expression in the given environment
 
@@ -124,7 +130,7 @@ class Base(SourceNode, ABC):
         result can be an instance of ``WDL.Value.Null`` which is distinct from None.
         """
         if isinstance(self, (Boolean, Int, Float)):
-            return self._eval(Env.Bindings())
+            return self._eval(Env.Bindings(), None)  # pyre-fixme
         return None
 
 
@@ -150,9 +156,7 @@ class Boolean(Base):
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         return Type.Boolean()
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Boolean:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Boolean:
         ""
         return Value.Boolean(self.value)
 
@@ -179,9 +183,7 @@ class Int(Base):
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         return Type.Int()
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Int:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Int:
         ""
         return Value.Int(self.value)
 
@@ -211,13 +213,38 @@ class Float(Base):
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         return Type.Float()
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Float:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Float:
         ""
         return Value.Float(self.value)
 
 
+class Null(Base):
+    """
+    WDL ``None`` literal
+
+    (called ``Null`` to avoid conflict with Python ``None``)
+    """
+
+    value: None
+    """
+    :type: None
+    """
+
+    def __init__(self, pos: SourcePosition) -> None:
+        super().__init__(pos)
+        self.value = None
+
+    def __str__(self):
+        return "None"
+
+    def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
+        return Type.Any(null=True)
+
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Null:
+        ""
+        return Value.Null()
+
+
 class Placeholder(Base):
     """Holds an expression interpolated within a string or command"""
 
@@ -239,6 +266,16 @@ class Placeholder(Base):
         self.options = options
         self.expr = expr
 
+        # preprocess expr to rewrite any Apply("_add") to Apply("_interpolation_add") for the
+        # special interpolation-only behavior of + for String? operands.
+        def rewrite_adds(ch: Base):
+            if isinstance(ch, Apply) and ch.function_name == "_add":
+                ch.function_name = "_interpolation_add"
+            for ch2 in ch.children:
+                rewrite_adds(ch2)
+
+        rewrite_adds(self.expr)
+
     def __str__(self):
         options = []
         for option in self.options:
@@ -250,18 +287,6 @@ class Placeholder(Base):
     def children(self) -> Iterable[SourceNode]:
         yield self.expr
 
-    def infer_type(
-        self,
-        type_env: Env.Bindings[Type.Base],
-        stdlib: "Optional[StdLib.Base]" = None,
-        check_quant: bool = True,
-    ) -> Base:
-        # override the + operator with the within-interpolation version which accepts String?
-        # operands and produces a String? result
-        stdlib = stdlib or StdLib.Base()
-        setattr(stdlib, "_add", StdLib.InterpolationAddOperator())
-        return super().infer_type(type_env, stdlib, check_quant)
-
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         if isinstance(self.expr.type, Type.Array):
             if "sep" not in self.options:
@@ -294,14 +319,8 @@ class Placeholder(Base):
                 )
         return Type.String()
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.String:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.String:
         ""
-        # override the + operator with the within-interpolation version which evaluates to None
-        # if either operand is None
-        stdlib = stdlib or StdLib.Base()
-        setattr(stdlib, "_add", StdLib.InterpolationAddOperator())
         v = self.expr.eval(env, stdlib)
         if isinstance(v, Value.Null):
             if "default" in self.options:
@@ -371,9 +390,7 @@ class String(Base):
         ""
         return super().typecheck(expected)  # pyre-ignore
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.String:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.String:
         ""
         ans = []
         for part in self.parts:
@@ -395,7 +412,7 @@ class String(Base):
     def literal(self) -> Optional[Value.Base]:
         if next((p for p in self.parts if not isinstance(p, str)), None):
             return None
-        return self._eval(Env.Bindings())
+        return self._eval(Env.Bindings(), None)  # pyre-fixme
 
 
 class Array(Base):
@@ -431,7 +448,7 @@ class Array(Base):
         item_type = Type.unify(
             [item.type for item in self.items], check_quant=self._check_quant, force_string=True
         )
-        if isinstance(item_type, Type.Any):
+        if isinstance(item_type, Type.Any) and not item_type.optional:
             raise Error.IndeterminateType(self, "unable to unify array item types")
         return Type.Array(item_type, optional=False, nonempty=True)
 
@@ -442,9 +459,7 @@ class Array(Base):
             return self
         return super().typecheck(expected)  # pyre-ignore
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Array:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Array:
         ""
         assert isinstance(self.type, Type.Array)
         return Value.Array(
@@ -499,9 +514,7 @@ class Pair(Base):
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         return Type.Pair(self.left.type, self.right.type)
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ""
         assert isinstance(self.type, Type.Pair)
         lv = self.left.eval(env, stdlib)
@@ -575,15 +588,18 @@ class Map(Base):
                     literal_keys = None
         return Type.Map((kty, vty), literal_keys=literal_keys)
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ""
         assert isinstance(self.type, Type.Map)
+        keystrs = set()
         eitems = []
         for k, v in self.items:
-            eitems.append((k.eval(env, stdlib), v.eval(env, stdlib)))
-        # TODO: complain of duplicate keys
+            ek = k.eval(env, stdlib)
+            sk = str(ek)
+            if sk in keystrs:
+                raise Error.EvalError(self, "duplicate keys in Map literal")
+            eitems.append((ek, v.eval(env, stdlib)))
+            keystrs.add(sk)
         return Value.Map(self.type.item_type, eitems)
 
     @property
@@ -613,13 +629,27 @@ class Struct(Base):
     can be coerced to a specific struct type during typechecking.
     """
 
-    def __init__(self, pos: SourcePosition, members: List[Tuple[str, Base]]):
+    struct_type_name: Optional[str]
+    """
+    :type: Optional[str]
+
+    In WDL 2.0+ each struct literal may specify the intended struct type name.
+    """
+
+    def __init__(
+        self,
+        pos: SourcePosition,
+        members: List[Tuple[str, Base]],
+        struct_type_name: Optional[str] = None,
+    ):
         super().__init__(pos)
         self.members = {}
         for (k, v) in members:
             if k in self.members:
                 raise Error.MultipleDefinitions(self.pos, "duplicate keys " + k)
             self.members[k] = v
+        self.struct_type_name = struct_type_name
+        assert struct_type_name is None or isinstance(struct_type_name, str), str(struct_type_name)
 
     def __str__(self):
         members = []
@@ -633,18 +663,36 @@ class Struct(Base):
         return self.members.values()
 
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
-        member_types = {}
-        for k, v in self.members.items():
-            member_types[k] = v.type
-        return Type.Object(member_types)
+        object_type = Type.Object({k: v.type for k, v in self.members.items()})
+        if not self.struct_type_name:
+            # pre-WDL 2.0: object literal with deferred typechecking
+            return object_type
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+        # resolve struct type
+        struct_type_members = None
+        if self._struct_types and self.struct_type_name in self._struct_types:
+            struct_type_members = self._struct_types[self.struct_type_name]
+        if struct_type_members is None:
+            raise Error.InvalidType(self, "Unknown type " + self.struct_type_name)
+
+        struct_type = Type.StructInstance(self.struct_type_name)
+        struct_type.members = struct_type_members
+
+        # typecheck members vs struct declaration
+        if not object_type.coerces(struct_type):
+            raise Error.StaticTypeMismatch(
+                self,
+                struct_type,
+                object_type,
+            )
+
+        return struct_type
+
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ans = {}
         for k, v in self.members.items():
             ans[k] = v.eval(env, stdlib)
-        assert isinstance(self.type, Type.Object)
+        assert isinstance(self.type, (Type.Object, Type.StructInstance))
         return Value.Struct(self.type, ans)
 
     @property
@@ -656,7 +704,7 @@ class Struct(Base):
                 ans[k] = vl
             else:
                 return None
-        assert isinstance(self.type, Type.Object)
+        assert isinstance(self.type, (Type.Object, Type.StructInstance))
         return Value.Struct(self.type, ans)
 
 
@@ -723,9 +771,7 @@ class IfThenElse(Base):
             )
         return ty
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ""
         if self.condition.eval(env, stdlib).expect(Type.Boolean()).value:
             ans = self.consequent.eval(env, stdlib)
@@ -785,9 +831,7 @@ class Ident(Base):
             self.referee = referee
         return ans
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ""
         return env[self.name]
 
@@ -818,9 +862,7 @@ class _LeftName(Base):
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
         raise NotImplementedError()
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         raise NotImplementedError()
 
     @property
@@ -898,6 +940,7 @@ class Get(Base):
                 raise Error.UnknownIdentifier(self)
         # attempt to typecheck expr, disambiguating whether it's an
         # intermediate value, a resolvable identifier, or neither
+        assert self._stdlib is not None
         try:
             self.expr.infer_type(type_env, self._stdlib, self._check_quant)
         except Error.UnknownIdentifier:
@@ -939,9 +982,7 @@ class Get(Base):
                 pass
         raise Error.NoSuchMember(self, self.member)
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         innard_value = self.expr.eval(env, stdlib)
         if not self.member:
             return innard_value
@@ -961,9 +1002,6 @@ class Get(Base):
         return ""
 
 
-_base_stdlib = None  # memorized instance of the default WDL.StdLib.Base()
-
-
 def _add_parentheses(arguments, parent_operator):
     """
     Add parentheses around arguments if necessary.
@@ -980,6 +1018,7 @@ def _add_parentheses(arguments, parent_operator):
         "_div": 7,
         "_rem": 7,
         "_add": 6,
+        "_interpolation_add": 6,
         "_sub": 6,
         "_lt": 5,
         "_lte": 5,
@@ -1024,20 +1063,27 @@ class Apply(Base):
         self.arguments = arguments
 
     def __str__(self):
-        func = getattr(StdLib.Base(), self.function_name)
         arguments = _add_parentheses(self.arguments, self.function_name)
-        if isinstance(func, StdLib._ArithmeticOperator):
-            return "{} {} {}".format(arguments[0], func.name, arguments[1])
-        elif isinstance(func, StdLib._ComparisonOperator):
-            return "{} {} {}".format(arguments[0], func.name, arguments[1])
-        elif isinstance(func, StdLib._At):
+        infix = {
+            "_mul": "*",
+            "_div": "/",
+            "_rem": "%",
+            "_add": "+",
+            "_interpolation_add": "+",
+            "_sub": "-",
+            "_lt": "<",
+            "_lte": "<=",
+            "_gt": ">",
+            "_gte": ">=",
+            "_eqeq": "==",
+            "_neq": "!=",
+            "_land": "&&",
+            "_lor": "||",
+        }
+        if self.function_name in infix:
+            return "{} {} {}".format(arguments[0], infix[self.function_name], arguments[1])
+        elif self.function_name == "_at":
             return "{}[{}]".format(arguments[0], arguments[1])
-        elif isinstance(func, StdLib._And):
-            return "{} && {}".format(arguments[0], arguments[1])
-        elif isinstance(func, StdLib._Or):
-            return "{} || {}".format(arguments[0], arguments[1])
-        elif self.function_name == "_rem":
-            return "{} % {}".format(arguments[0], arguments[1])
         elif self.function_name == "_negate":
             return "!{}".format(arguments[0])
         else:
@@ -1050,24 +1096,15 @@ class Apply(Base):
 
     def _infer_type(self, type_env: Env.Bindings[Type.Base]) -> Type.Base:
 
-        global _base_stdlib
-        if not _base_stdlib:
-            _base_stdlib = StdLib.Base()
-        f = getattr(self._stdlib or _base_stdlib, self.function_name, None)
+        f = getattr(self._stdlib, self.function_name, None)
         if not f:
             raise Error.NoSuchFunction(self, self.function_name) from None
         assert isinstance(f, StdLib.Function)
         return f.infer_type(self)
 
-    def _eval(
-        self, env: Env.Bindings[Value.Base], stdlib: "Optional[StdLib.Base]" = None
-    ) -> Value.Base:
+    def _eval(self, env: Env.Bindings[Value.Base], stdlib: StdLib.Base) -> Value.Base:
         ""
 
-        global _base_stdlib
-        if not _base_stdlib:
-            _base_stdlib = StdLib.Base()
-        stdlib = stdlib or _base_stdlib
         f = getattr(stdlib, self.function_name, None)
         assert isinstance(f, StdLib.Function)
         return f(self, env, stdlib)
