@@ -14,6 +14,7 @@ import fcntl
 import subprocess
 import shutil
 import urllib
+import hashlib
 from time import sleep
 from datetime import datetime
 from contextlib import contextmanager, AbstractContextManager
@@ -174,6 +175,23 @@ def write_atomic(contents: str, filename: str, end: str = "\n") -> None:
     os.rename(tn, filename)
 
 
+@export
+def rmtree_atomic(path: str) -> None:
+    """
+    Recursively delete a directory (or single file) after first renaming it to a temporary name in
+    the same parent directory. The atomic rename step ensures a "partial" directory won't be left
+    behind in its original location, should anything go wrong whilst deleting its contents.
+    """
+    path = os.path.abspath(path)
+    assert path and path.strip("/")
+    tmp_path = os.path.join(
+        os.path.dirname(path), "._rmtree_atomic_" + hashlib.sha256(path.encode("utf-8")).hexdigest()
+    )
+    shutil.rmtree(tmp_path, ignore_errors=True)
+    os.renames(path, tmp_path)
+    shutil.rmtree(tmp_path)
+
+
 @export
 def write_values_json(
     values_env: "Env.Bindings[Value.Base]", filename: str, namespace: str = ""  # noqa
@@ -560,6 +578,29 @@ def parse_byte_size(s: str) -> int:
     return int(N)
 
 
+@export
+def pathsize(path: str) -> int:
+    """
+    get byte size of file, or total size of all files under directory & subdirectories (symlinks
+    excluded)
+    """
+
+    if not os.path.isdir(path):
+        return os.path.getsize(path)
+
+    def raiser(exc: OSError):
+        raise exc
+
+    ans = 0
+    for root, subdirs, files in os.walk(path, onerror=raiser, followlinks=False):
+        for fn in files:
+            fn = os.path.join(root, fn)
+            if not os.path.islink(fn):
+                ans += os.path.getsize(fn)
+
+    return ans
+
+
 def splitall(path: str) -> List[str]:
     """
     https://www.oreilly.com/library/view/python-cookbook/0596001673/ch04s16.html
@@ -598,11 +639,15 @@ def chmod_R_plus(path: str, file_bits: int = 0, dir_bits: int = 0) -> None:
 
     def do1(path1: str, bits: int) -> None:
         assert 0 <= bits < 0o10000
-        if path_really_within(path1, path):
+        if not os.path.islink(path1) and path_really_within(path1, path):
             os.chmod(path1, (os.stat(path1).st_mode & 0o7777) | bits)
 
+    def raiser(exc: OSError):
+        raise exc
+
     if os.path.isdir(path):
-        for root, subdirs, files in os.walk(path, followlinks=False):
+        do1(path, dir_bits)
+        for root, subdirs, files in os.walk(path, onerror=raiser, followlinks=False):
             for dn in subdirs:
                 do1(os.path.join(root, dn), dir_bits)
             for fn in files:
@@ -712,7 +757,7 @@ class FlockHolder(AbstractContextManager):
     """
 
     _lock: threading.Lock
-    _flocks: Dict[str, Tuple[IO[Any], bool]]
+    _flocks: Dict[str, Tuple[int, bool]]
     _entries: int
     _logger: logging.Logger
 
@@ -735,10 +780,10 @@ class FlockHolder(AbstractContextManager):
         if self._entries == 0:
             exn = None
             with self._lock:
-                for fn, (fh, exclusive) in self._flocks.items():
+                for fn, (fd, exclusive) in self._flocks.items():
                     self._logger.debug(StructuredLogMessage("close", file=fn, exclusive=exclusive))
                     try:
-                        fh.close()
+                        os.close(fd)
                     except Exception as exn2:
                         exn = exn or exn2
                 self._flocks = {}
@@ -748,21 +793,21 @@ class FlockHolder(AbstractContextManager):
     def __del__(self) -> None:
         assert self._entries == 0 and not self._flocks, "FlockHolder context was not exited"
 
-    def flock(  # pyre-fixme
+    def flock(
         self,
         filename: str,
-        mode: str = "",
+        mode: Optional[int] = None,
         exclusive: bool = False,
         wait: bool = False,
         update_atime: bool = False,
-    ) -> IO[Any]:
+    ) -> int:
         """
-        Open a file and an advisory lock on it. The file is closed and the lock released upon exit
-        of the outermost context. Returns the open file, which the caller shouldn't close (this is
-        taken care of).
+        Open a file/directory and an advisory lock on it. The file is closed and the lock released
+        upon exit of the outermost context. Returns the open file descriptor, which the caller
+        shouldn't close (this is taken care of).
 
-        :param filename: file to open & lock
-        :param mode: open() mode, default: "r+b" if exclusive else "rb"
+        :param filename: file/directory to open & lock
+        :param mode: os.open() mode flags, default: OS.O_RDWR if exclusive else os.O_RDONLY
         :param exclusive: True to open an exclusive lock (default: shared lock)
         :param wait: True to wait as long as needed to obtain the lock, otherwise (default) raise
                      OSError if the lock isn't available immediately. Self-deadlock is possible;
@@ -783,7 +828,10 @@ class FlockHolder(AbstractContextManager):
                         )
                     )
                     return self._flocks[realfilename][0]
-                openfile = open(realfilename, mode or ("r+b" if exclusive else "rb"))
+                openfile = os.open(
+                    realfilename,
+                    mode if mode is not None else (os.O_RDWR if exclusive else os.O_RDONLY),
+                )
                 try:
                     op = fcntl.LOCK_EX if exclusive else fcntl.LOCK_SH
                     if not wait:
@@ -800,11 +848,9 @@ class FlockHolder(AbstractContextManager):
                     fcntl.flock(openfile, op)
                     # the flock will release whenever we ultimately openfile.close()
 
-                    file_st = os.stat(openfile.fileno())
+                    file_st = os.stat(openfile)
                     if update_atime:
-                        os.utime(
-                            openfile.fileno(), ns=(int(time.time() * 1e9), file_st.st_mtime_ns)
-                        )
+                        os.utime(openfile, ns=(int(time.time() * 1e9), file_st.st_mtime_ns))
 
                     # Even if all concurrent processes obey the advisory flocks, the filename link
                     # could have been replaced or removed in the duration between our open() and
@@ -832,9 +878,9 @@ class FlockHolder(AbstractContextManager):
                         self._flocks[realfilename] = (openfile, exclusive)
                         return openfile
                 except:
-                    openfile.close()
+                    os.close(openfile)
                     raise
-                openfile.close()
+                os.close(openfile)  # NOT finally -- for next while-loop iteration
 
 
 @export
