@@ -9,8 +9,10 @@ import json
 import itertools
 import os
 import logging
+import shutil
+import base64
 from pathlib import Path
-from typing import Dict, Any, Optional, List
+from typing import Dict, Any, Optional, List, Iterable, Union
 from contextlib import AbstractContextManager
 from urllib.parse import urlparse, urlunparse
 from fnmatch import fnmatchcase
@@ -24,6 +26,7 @@ from .._util import (
     StructuredLogMessage as _,
     FlockHolder,
     write_atomic,
+    rmtree_atomic,
 )
 
 
@@ -36,6 +39,7 @@ class CallCache(AbstractContextManager):
     # the course of the current workflow run, but not eligible for persistent caching in future
     # runs; we just want to remember them for potential reuse later in the current run.
     _workflow_downloads: Dict[str, str]
+    _workflow_directory_downloads: Dict[str, str]
     _lock: Lock
 
     def __init__(self, cfg: config.Loader, logger: logging.Logger):
@@ -43,6 +47,7 @@ class CallCache(AbstractContextManager):
         self._logger = logger.getChild("CallCache")
         self._flocker = FlockHolder(self._logger)
         self._workflow_downloads = {}
+        self._workflow_directory_downloads = {}
         self._lock = Lock()
         self.call_cache_dir = cfg["call_cache"]["dir"]
 
@@ -65,13 +70,15 @@ class CallCache(AbstractContextManager):
         from .. import values_to_json
 
         json_inputs = json.dumps(values_to_json(inputs), sort_keys=True).encode("utf-8")
-        return hashlib.sha256(json_inputs).hexdigest()
+        sha256 = hashlib.sha256(json_inputs).digest()
+        return base64.b32encode(sha256[:20]).decode().lower()
 
     def get_digest_for_task(self, task):
         doc = getattr(task, "parent", None)
         assert isinstance(doc, Document)
-        task_string = _describe_task(doc, task)
-        return hashlib.sha256(task_string.encode("utf-8")).hexdigest()
+        task_string = _describe_task(doc, task).encode("utf-8")
+        sha256 = hashlib.sha256(task_string).digest()
+        return base64.b32encode(sha256[:20]).decode().lower()
 
     def get(
         self, key: str, output_types: Env.Bindings[Type.Base], inputs: Env.Bindings[Value.Base]
@@ -84,7 +91,7 @@ class CallCache(AbstractContextManager):
         from .. import values_from_json
 
         file_path = os.path.join(self.call_cache_dir, f"{key}.json")
-        file_coherence_checker = FileCoherence(self._logger)
+        file_coherence_checker = FileCoherence(self._cfg, self._logger)
 
         if not self._cfg["call_cache"].get_bool("get"):
             return None
@@ -101,15 +108,20 @@ class CallCache(AbstractContextManager):
             )
         if cache:
             self._logger.notice(_("call cache hit", cache_path=file_path))  # pyre-fixme
-            file_list = []
-            # check output and input files
+            file_list = set()
+            dir_list = set()
+            # check output and input file timestamps
 
-            def get_files(file):
-                file_list.append(file)
+            def get_files(v: Union[Value.File, Value.Directory]):
+                if isinstance(v, Value.File):
+                    file_list.add(v.value)
+                else:
+                    assert isinstance(v, Value.Directory)
+                    dir_list.add(v.value)
 
-            Value.rewrite_env_files(cache, get_files)
-            Value.rewrite_env_files(inputs, get_files)
-            if file_coherence_checker.check_files(file_path, file_list):
+            Value.rewrite_env_paths(cache, get_files)
+            Value.rewrite_env_paths(inputs, get_files)
+            if file_coherence_checker.check_files(file_path, file_list, dir_list):
                 return cache
         return None
 
@@ -134,13 +146,13 @@ class CallCache(AbstractContextManager):
     # specialized caching logic for file downloads (not sensitive to the downloader task details,
     # and looked up in URI-derived folder structure instead of sqlite db)
 
-    def download_path(self, uri: str) -> Optional[str]:
+    def download_path(self, uri: str, directory: bool = False) -> Optional[str]:
         """
         Based on the input download uri, compute the local file path at which the cached copy
         should exist (or None if the uri is not cacheable)
         """
         # check if URI is properly formatted & normalize
-        parts = urlparse(uri)
+        parts = urlparse(uri.rstrip("/"))
         if (
             parts.scheme
             and parts.netloc
@@ -158,14 +170,18 @@ class CallCache(AbstractContextManager):
             ):
                 (dn, fn) = os.path.split(parts.path)
                 if fn:
-                    # formulate path
+                    # formulate local subdirectory of the cache directory in which to put the
+                    # cached item, manipulating the URI path to ensure consistent local nesting
+                    # depth (that's assumed by clean_download_cache.sh when it's looking for items
+                    # to clean up)
                     dn = dn.strip("/")
                     if dn:
                         dn = dn.replace("_", "__")
                         dn = dn.replace("/", "_")
+                    dn = "_" + dn
                     return os.path.join(
                         self._cfg["download_cache"]["dir"],
-                        "files",
+                        ("dirs" if directory else "files"),
                         parts.scheme,
                         parts.netloc,
                         dn,
@@ -173,17 +189,27 @@ class CallCache(AbstractContextManager):
                     )
         return None
 
-    def get_download(self, uri: str, logger: Optional[logging.Logger] = None) -> Optional[str]:
+    def get_download(
+        self, uri: str, directory: bool = False, logger: Optional[logging.Logger] = None
+    ) -> Optional[str]:
         """
         Return filename of the cached download of uri, if available. If so then opens a shared
-        flock on the local file, which will remain for the life of the CallCache object.
+        flock on the local file/directory, which will remain for the life of the CallCache object.
         """
+        if directory:
+            uri = uri.rstrip("/")
         with self._lock:
-            if uri in self._workflow_downloads:
+            if directory and uri in self._workflow_directory_downloads:
+                return self._workflow_directory_downloads[uri]
+            elif not directory and uri in self._workflow_downloads:
                 return self._workflow_downloads[uri]
         logger = logger.getChild("CallCache") if logger else self._logger
-        p = self.download_path(uri)
-        if not (self._cfg["download_cache"].get_bool("get") and p and os.path.isfile(p)):
+        p = self.download_path(uri, directory=directory)
+        if not (
+            self._cfg["download_cache"].get_bool("get")
+            and p
+            and ((directory and os.path.isdir(p)) or (not directory and os.path.isfile(p)))
+        ):
             logger.debug(_("no download cache hit", uri=uri, cache_path=p))
             return None
         try:
@@ -201,26 +227,34 @@ class CallCache(AbstractContextManager):
             )
             return None
 
-    def put_download(self, uri: str, filename: str, logger: Optional[logging.Logger] = None) -> str:
+    def put_download(
+        self,
+        uri: str,
+        filename: str,
+        directory: bool = False,
+        logger: Optional[logging.Logger] = None,
+    ) -> str:
         """
         Move the downloaded file to the cache location & return the new path; or if the uri isn't
         cacheable, return the given path.
         """
+        if directory:
+            uri = uri.rstrip("/")
         logger = logger.getChild("CallCache") if logger else self._logger
         ans = filename
-        p = self.download_cacheable(uri)
+        p = self.download_cacheable(uri, directory=directory)
         if p:
             # if a file at the cache location has appeared whilst we were downloading, replace it
             # iff we can exclusive-flock it
             with FlockHolder(logger) as replace_flock:
                 try:
-                    replace_flock.flock(p, mode="rb", exclusive=True)
+                    replace_flock.flock(p, mode=os.O_RDONLY, exclusive=True)
                 except FileNotFoundError:
                     pass
                 except OSError:
                     logger.warning(
                         _(
-                            "existing cached file in use; leaving downloaded in-place",
+                            "existing cache entry in use; leaving downloaded in-place",
                             uri=uri,
                             downloaded=filename,
                             cache_path=p,
@@ -228,20 +262,25 @@ class CallCache(AbstractContextManager):
                     )
                     p = None
                 if p:
-                    os.makedirs(os.path.dirname(p), exist_ok=True)
-                    os.rename(filename, p)
+                    if directory and os.path.isdir(p):
+                        rmtree_atomic(p)
+                    os.renames(filename, p)
+                    # the renames() op should be atomic, because the download operation should have
+                    # been run under the cache directory (download.py:run_cached)
                     logger.info(_("stored in download cache", uri=uri, cache_path=p))
                     ans = p
         if not p:
             with self._lock:
-                self._workflow_downloads[uri] = ans
+                (self._workflow_directory_downloads if directory else self._workflow_downloads)[
+                    uri
+                ] = ans
         self.flock(ans)
         return ans
 
-    def download_cacheable(self, uri: str) -> Optional[str]:
+    def download_cacheable(self, uri: str, directory: bool = False) -> Optional[str]:
         if not self._cfg["download_cache"].get_bool("put"):
             return None
-        return self.download_path(uri)
+        return self.download_path(uri, directory=directory)
 
     def flock(self, filename: str, exclusive: bool = False) -> None:
         self._flocker.flock(filename, update_atime=True, exclusive=exclusive)
@@ -348,24 +387,47 @@ class FileCoherence(abc.ABC):
     for cache and referenced files) for files stored locally.
     """
 
+    _cfg: config.Loader
     _logger: logging.Logger
 
-    def __init__(self, logger):
+    def __init__(self, cfg: config.Loader, logger: logging.Logger):
+        self._cfg = cfg
         self._logger = logger
         self.cache_file_modification_time = 0.0
 
-    def check_files(self, cache_file_path: str, files: list) -> bool:
+        # working around circular import
+        from .download import able as downloadable
+
+        self._downloadable = downloadable
+
+    def check_files(self, cache_file_path: str, files: Iterable[str], dirs: Iterable[str]) -> bool:
         if self.cache_file_modification_time == 0.0:
             self.cache_file_modification_time = self.get_last_modified_time(cache_file_path)
-        for file_path in files:
+
+        def raiser(exc):
+            raise exc
+
+        for directory, path in itertools.chain(
+            ((False, f) for f in files), ((True, d) for d in dirs)
+        ):
             try:
-                self.check_cache_younger_than_file(output_file_path=file_path)
-            except (FileNotFoundError, CacheOutputFileAgeError):
+                if not self._downloadable(self._cfg, path):
+                    self.check_cache_younger_than_file(path)
+                    if directory:
+                        # check everything in directory
+                        for root, subdirs, subfiles in os.walk(
+                            path, onerror=raiser, followlinks=False
+                        ):
+                            for subdir in subdirs:
+                                self.check_cache_younger_than_file(os.path.join(root, subdir))
+                            for fn in subfiles:
+                                self.check_cache_younger_than_file(os.path.join(root, fn))
+            except (FileNotFoundError, NotADirectoryError, CacheOutputFileAgeError):
                 self._logger.warning(
                     _(
-                        "cache entry invalid due to deleted or modified file",
-                        cache_path=file_path,
-                        file_changed=file_path,
+                        "cache entry invalid due to deleted or modified file/directory",
+                        cache_path=path,
+                        file_changed=path,
                     )
                 )
                 try:
@@ -374,7 +436,7 @@ class FileCoherence(abc.ABC):
                     self._logger.warning(
                         _(
                             "unable to delete invalidated cache entry",
-                            cache_path=file_path,
+                            cache_path=path,
                             error=str(exn),
                         )
                     )
@@ -382,14 +444,16 @@ class FileCoherence(abc.ABC):
         return True
 
     def get_last_modified_time(self, file_path: str) -> float:
-        # returned as seconds since epoch
-        file_modification_time = os.path.getmtime(file_path)
-        sym_link_modification_time = os.lstat(file_path).st_mtime
-
-        return max(file_modification_time, sym_link_modification_time)
+        # max mtime of hardlink & symlink pointing to it (if applicable)
+        return max(
+            os.stat(file_path, follow_symlinks=False).st_mtime_ns,
+            os.stat(file_path, follow_symlinks=True).st_mtime_ns,
+        )
 
     def check_cache_younger_than_file(self, output_file_path: str) -> bool:
         output_file_modification_time = self.get_last_modified_time(output_file_path)
+        # self._logger.debug(_("check_cache_younger_than_file", path=output_file_path,
+        # mtime=output_file_modification_time/1e9, cache_mtime=self.cache_file_modification_time/1e9))
         if self.cache_file_modification_time >= output_file_modification_time:
             return True
         else:
