@@ -71,10 +71,16 @@ class DownloadFailed(_RuntimeError):
         _statusbar.abort()
 
 
-class RunFailed(_RuntimeError):
+class CacheOutputFileAgeError(_RuntimeError):
+    """
+    Raised when an output file has been modified more recently than the cache file that refers to it
     """
 
-    """
+    pass
+
+
+class RunFailed(_RuntimeError):
+    """"""
 
     exe: Union[_Task, _Workflow]
     run_id: str
