@@ -280,7 +280,11 @@ class TestStdLib(unittest.TestCase):
         """)
         self.assertEqual(outputs["ai"], [1, 2, 3, 1, 21, 22])
         self.assertEqual(outputs["af"], ["/tmp/X.txt", "/tmp/Y.txt", "/tmp/Z.txt"])
-        self.assertEqual(outputs["ap"], [[0.1, "mouse"], [3, "cat"], [15, "dog"]])
+        self.assertEqual(outputs["ap"], [
+            {"left": 0.1, "right": "mouse"},
+            {"left": 3, "right": "cat"},
+            {"left": 15, "right": "dog"}
+        ])
 
     def test_size(self):
         with open(os.path.join(self._dir, "alyssa.txt"), "w") as outfile:
@@ -696,8 +700,19 @@ class TestStdLib(unittest.TestCase):
             }
         }
         """)
-        self.assertEqual(outputs["zipped"], [[1, "a"], [2, "b"], [3, "c"]])
-        self.assertEqual(outputs["crossed"], [[1, "d"], [1, "e"], [2, "d"], [2, "e"], [3, "d"], [3, "e"]])
+        self.assertEqual(outputs["zipped"], [
+            {"left": 1, "right": "a"},
+            {"left": 2, "right": "b"},
+            {"left": 3, "right": "c"}
+        ])
+        self.assertEqual(outputs["crossed"], [
+            {"left": 1, "right": "d"},
+            {"left": 1, "right": "e"},
+            {"left": 2, "right": "d"},
+            {"left": 2, "right": "e"},
+            {"left": 3, "right": "d"},
+            {"left": 3, "right": "e"}
+        ])
 
         outputs = self._test_task(R"""
         version 1.0
@@ -816,7 +831,7 @@ class TestStdLib(unittest.TestCase):
 
     def test_quote(self):
         outputs = self._test_task(R"""
-        version 1.0
+        version development
         task test_quote {
             command {}
             output {
@@ -832,7 +847,7 @@ class TestStdLib(unittest.TestCase):
         })
 
         outputs = self._test_task(R"""
-        version 1.0
+        version development
         task test_quote {
             command {}
             output {
@@ -849,8 +864,8 @@ class TestStdLib(unittest.TestCase):
         })
 
         # Check invalid type does not work
-        outputs = self._test_task(R"""
-        version 1.0
+        self._test_task(R"""
+        version development
         task test_quote {
             command {}
             output {
@@ -860,10 +875,22 @@ class TestStdLib(unittest.TestCase):
         }
         """,expected_exception=WDL.Error.StaticTypeMismatch)
 
+        # check unavailable in WDL draft-2 and 1.0
+        self._test_task(R"""
+        version 1.0
+        task test_quote {
+            command {}
+            output {
+                Array[Int] arguments = [1,2,3]
+                Array[String] quoted_args = quote(arguments) # ["\"1\"","\"2\"","\"3\""]
+            }
+        }
+        """, expected_exception=WDL.Error.NoSuchFunction)
+
 
     def test_squote(self):
         outputs = self._test_task(R"""
-        version 1.0
+        version development
         task test_squote {
             command {}
             output {
@@ -879,7 +906,7 @@ class TestStdLib(unittest.TestCase):
         })
 
         outputs = self._test_task(R"""
-        version 1.0
+        version development
         task test_squote {
             command {}
             output {
@@ -897,7 +924,7 @@ class TestStdLib(unittest.TestCase):
 
         # Check invalid type does not work
         outputs = self._test_task(R"""
-        version 1.0
+        version development
         task test_squote {
             command {}
             output {
@@ -906,3 +933,112 @@ class TestStdLib(unittest.TestCase):
             }
         }
         """,expected_exception=WDL.Error.StaticTypeMismatch)
+
+    def test_keys(self):
+        outputs = self._test_task(R"""
+        version development
+        task test_keys {
+            input {
+                Map[String,String] m1 = {"a": "b", "c": "d"}
+                Map[Int,Boolean] m2 = {1: true, -1: false}
+                Map[Int,Float]? m3
+            }
+            command {}
+            output {
+                Array[String] k1 = keys(m1)
+                Array[Int] k2 = keys(m2)
+                Array[Boolean] k4 = keys({})
+                Array[Pair[Int,Boolean]] k5 = keys({(1,false): "foo", (3,true): "bar"})
+            }
+        }
+        """)
+        self.assertEqual(outputs["k1"], ["a", "c"])
+        self.assertEqual(outputs["k2"], [1,-1])
+        self.assertEqual(outputs["k4"], [])
+        self.assertEqual(outputs["k5"], [{"left": 1, "right": False}, {"left": 3, "right": True}])
+
+        with self.assertRaises(WDL.Error.StaticTypeMismatch):
+            self._test_task(R"""
+            version development
+            task test_keys {
+                input {
+                    Array[Int] a = keys([1,2,3])
+                }
+                command {}
+                output {}
+            }
+            """)
+
+    def test_map_pairs(self):
+        outputs = self._test_task(R"""
+        version development
+        task test_map_pairs {
+            input {
+                Array[Pair[String,Int]] x = [("b", 1), ("a", 2), ("c", 3)]
+                Array[Pair[String,Pair[String,String]]] y = [("a", ("a_1.bam", "a_1.bai")), ("b", ("b.bam", "b.bai")), ("a", ("a_2.bam", "a_2.bai"))]
+                Array[Pair[String,Pair[String,String]]] y2 = [("a", ("a.bam", "a.bai")), ("b", ("b.bam", "b.bai"))]
+            }
+
+            Map[String,Int] xmap = as_map(x)
+            Map[String,Pair[String,String]] ymap = as_map(y2)
+
+            command {}
+
+            output {
+                Map[String,Int] xmap_out = xmap
+                Map[String,Pair[String,String]] ymap_out = ymap
+                Map[String,Array[Int]] xmulti = collect_by_key(x)
+                Map[String,Array[Pair[String,String]]] ymulti = collect_by_key(y)
+                Array[Pair[String,Int]] x_roundtrip = as_pairs(xmap)
+                Array[Pair[String,Pair[String,String]]] y_roundtrip = as_pairs(ymap)
+            }
+        }
+        """)
+        self.assertEqual(outputs["xmap_out"], {"b": 1, "a": 2, "c": 3})
+        self.assertEqual(outputs["ymap_out"], {"a": {"left": "a.bam", "right": "a.bai"}, "b": {"left": "b.bam", "right": "b.bai"}})
+        self.assertEqual(outputs["xmulti"], {"b": [1], "a": [2], "c": [3]})
+        self.assertEqual(outputs["ymulti"], {"a": [{"left": "a_1.bam", "right": "a_1.bai"}, {"left": "a_2.bam", "right": "a_2.bai"}], "b": [{"left": "b.bam", "right": "b.bai"}]})
+        self.assertEqual(outputs["x_roundtrip"], [{"left": "b", "right": 1}, {"left": "a", "right": 2}, {"left": "c", "right": 3}])
+        self.assertEqual(outputs["y_roundtrip"], [{"left": "a", "right": {"left": "a.bam", "right": "a.bai"}}, {"left": "b", "right": {"left": "b.bam", "right": "b.bai"}}])
+
+        with self.assertRaises(WDL.Error.StaticTypeMismatch):
+            self._test_task(R"""
+            version development
+            task test_keys {
+                input {
+                    Map[String,Int]? optmap
+                }
+                command {}
+                output {
+                    Array[Pair[String,Int]] x = as_pairs(optmap)
+                }
+            }
+            """)
+
+        with self.assertRaises(WDL.Error.StaticTypeMismatch):
+            self._test_task(R"""
+            version development
+            task test_keys {
+                input {
+                    Array[Array[Int]] a2
+                }
+                command {}
+                output {
+                    Map[Int,Int] x = collect_by_key(a2)
+                }
+            }
+            """)
+
+        with self.assertRaisesRegex(WDL.Error.EvalError, "duplicate"):
+            outputs = self._test_task(R"""
+            version development
+            task test_map_pairs {
+                input {
+                    Array[Pair[String,Pair[String,String]]] y = [("a", ("a_1.bam", "a_1.bai")), ("b", ("b.bam", "b.bai")), ("a", ("a_2.bam", "a_2.bai"))]
+                }
+                command {}
+                output {
+                    Map[String,Pair[String,String]] ymap = as_map(y)
+                }
+            }
+            """)
